from cuqi.geometry import Geometry, MappedGeometry, _WrappedGeometry
import numpy as np
import matplotlib.pyplot as plt
import dolfin as dl
import ufl

__all__ = [
    'FEniCSContinuous',
    'FEniCSMappedGeometry',
    'MaternKLExpansion'
]

class FEniCSContinuous(Geometry):

    def __init__(self, function_space, labels = ['x', 'y']):
        self.function_space = function_space
        if self.physical_dim >2:
            raise NotImplementedError("'FenicsContinuous' object does not support 3D meshes yet. 'mesh' needs to be a 1D or 2D mesh.")
        self.labels = labels

    @property
    def physical_dim(self):
        return self.function_space.mesh().geometry().dim()  

    @property
    def mesh(self):
        return self.function_space.mesh()

    @property
    def par_shape(self):
        return (self.function_space.dim(),)

    def par2fun(self,par):
        """The parameter to function map used to map parameters to function values in e.g. plotting."""
        par = self._process_values(par)
        Ns = par.shape[-1]
        fun_list = []
        for idx in range(Ns):
            fun = dl.Function(self.function_space)
            fun.vector().zero()
            fun.vector().set_local(par[...,idx])
            fun_list.append(fun)

        if len(fun_list) == 1:
            return fun_list[0]
        else:
            return fun_list

    def fun2par(self,fun):
        """ Map the function values (FEniCS object) to the corresponding parameters (ndarray)."""
        return fun.vector().get_local()

    def gradient(self, direction, wrt=None, is_direction_par=False, is_wrt_par=True):
        """ Computes the gradient of the par2fun map with respect to the parameters in the direction `direction` evaluated at the point `wrt`"""
        if is_direction_par:
            return direction
        else:
            return self.fun2par(direction)

    def _plot(self,values,subplots=True,**kwargs):
        """
        Overrides :meth:`cuqi.geometry.Geometry.plot`. See :meth:`cuqi.geometry.Geometry.plot` for description  and definition of the parameter `values`.
        
        Parameters
        -----------
        kwargs : keyword arguments
            keyword arguments which the function :meth:`dolfin.plot` normally takes.
        """
        if isinstance(values, dl.function.function.Function):
            Ns = 1
            values = [values]
        elif hasattr(values,'__len__'): 
            Ns = len(values)
        subplot_ids = self._create_subplot_list(Ns,subplots=subplots)

        ims = []
        for rows,cols,subplot_id in subplot_ids:
            fun = values[subplot_id-1]
            if subplots:
                plt.subplot(rows,cols,subplot_id); 
            ims.append(dl.plot(fun, **kwargs))

        self._plot_config(subplots) 
        return ims

    def _process_values(self, values):
        if isinstance(values, dl.function.function.Function):
            return [values]

        elif len(values.shape) == 1:
            values = values[..., np.newaxis]
        
        return values
    
    def _plot_config(self, subplot):
        if self.labels is not None:
            if subplot == False:
                plt.gca().set_xlabel(self.labels[0])
                if self.physical_dim == 2: plt.gca().set_ylabel(self.labels[1]) 
            else:
                for i, axis in enumerate(plt.gcf().axes):
                    axis.set_xlabel(self.labels[0])
                    if self.physical_dim == 2: axis.set_ylabel(self.labels[1])


class FEniCSMappedGeometry(MappedGeometry):
    """
    """
    def par2fun(self,p):
        funvals = self.geometry.par2fun(p)
        if isinstance(funvals, dl.function.function.Function):
            funvals = [funvals]
        mapped_value_list = []
        for idx in range(len(funvals)):
            mapped_value = self.map(funvals[idx]) 
            if isinstance(mapped_value, ufl.algebra.Operator):
                mapped_value_list.append(dl.project(mapped_value, self.geometry.function_space))
            elif isinstance(mapped_value,dl.function.function.Function):
                mapped_value_list.append(mapped_value)
            else:
                raise ValueError(f"'{self.__class__.__name__}.map' should return 'ufl.algebra.Operator'")
            
        if len(mapped_value_list) == 1:
            return mapped_value_list[0]
        else:
            return mapped_value_list
    
    def fun2par(self,f):
        raise NotImplementedError


class MaternKLExpansion(_WrappedGeometry):
    """A geometry class that builds spectral representation of Matern covariance operator on the given input geometry. We create the representation using the stochastic partial differential operator, equation (15) in (Roininen, Huttunen and Lasanen, 2014). Zero Neumann boundary conditions are assumed for the stochastic partial differential equation (SPDE) and the default value of the smoothness parameter :math:`\\nu` is set to 0.5. To generate Matern field realizations, the method :meth:`par2field` is used. The input `p` of this method need to be an `n=dim` i.i.d random variables that follow a normal distribution. 

    For more details about the formulation of the SPDE see: Roininen, L., Huttunen, J. M., & Lasanen, S. (2014). Whittle-Matérn priors for Bayesian statistical inversion with applications in electrical impedance tomography. Inverse Problems & Imaging, 8(2), 561.

    Parameters
    -----------
    geometry : cuqi.fenics.geometry.Geometry
        An input geometry on which the Matern field representation is built (the geometry must have a mesh attribute)

    length_scale : float
        Length scale parameter (controls correlation length)

    num_terms: int
        Number of expansion terms to represent the Matern field realization

    nu : float, default 0.5 
        Smoothness parameter of the Matern field, must be greater then
        zero.

    boundary_conditions : str
        Boundary conditions for the SPDE. Currently 'Neumann' for zero flux, and 'zero' for zero Dirichlet boundary conditions are supported.

    normalize : bool, default True
        If True, the Matern field expansion modes are normalized to have a unit norm.

    Example
    -------
    .. code-block:: python

        import numpy as np
        import matplotlib.pyplot as plt
        from cuqi.fenics.geometry import MaternKLExpansion, FEniCSContinuous
        from cuqi.distribution import Gaussian
        import dolfin as dl
        
        mesh = dl.UnitSquareMesh(20,20)
        V = dl.FunctionSpace(mesh, 'CG', 1)
        geometry = FEniCSContinuous(V)
        MaternGeometry = MaternKLExpansion(geometry, 
                                        length_scale = .2,
                                        num_terms=128)
        
        MaternField = Gaussian(mean=np.zeros(MaternGeometry.dim),
                        cov=np.eye(MaternGeometry.dim),
                        geometry= MaternGeometry)
        
        samples = MaternField.sample()
        samples.plot()

    """

    def __init__(self, geometry, length_scale, num_terms, nu=0.5, boundary_conditions='Neumann', normalize=True): 

        if nu <= 0:
            raise ValueError("Smoothness parameter nu must be positive")
        
        if not isinstance(geometry, (FEniCSMappedGeometry, FEniCSContinuous)):
            raise ValueError("Matern KL expansion is only implemented "+ 
                             "for cuqipy_fenics geometries")
        super().__init__(geometry)
        if not hasattr(geometry, 'mesh'):
            raise NotImplementedError
        self._length_scale = length_scale
        self._nu = nu
        self._num_terms = num_terms
        self._eig_val = None
        self._eig_vec = None
        self._boundary_conditions = boundary_conditions
        self._normalize = normalize

    @property
    def par_shape(self):
        return (self.num_terms,)

    @property
    def length_scale(self):
        return self._length_scale

    @property
    def nu(self):
        return self._nu

    @property
    def num_terms(self):
        return self._num_terms

    @property
    def eig_val(self):
        return self._eig_val

    @property
    def eig_vec(self):
        return self._eig_vec

    @property
    def boundary_conditions(self):
        return self._boundary_conditions

    @property
    def normalize(self):
        return self._normalize
    
    @property
    def physical_dim(self):
        """Returns the physical dimension of the geometry, e.g. 1, 2 or 3"""
        return self.geometry.physical_dim

    def __repr__(self) -> str:
        return "{} on {}".format(self.__class__.__name__,self.geometry.__repr__())

    def par2fun(self,p):
        return self.geometry.par2fun(self.par2field(p))

    def gradient(self, direction, wrt):
        direction = self.geometry.gradient(direction, wrt)
        return np.diag( np.sqrt(self.eig_val)).T@self.eig_vec.T@direction
        
    def par2field(self, p):
        """Applies linear transformation of the parameters p to
        generate a realization of the Matern field (given that p is a
        sample of `n=dim` i.i.d random variables that follow a normal
        distribution)"""

        if self._eig_vec is None and self._eig_val is None:
            self._build_basis() 
	   
        p = self._process_values(p)
        Ns = p.shape[-1]
        field_list = np.empty((self.geometry.par_dim,Ns))

        nu = self.nu
        d = self.physical_dim
        for idx in range(Ns):
            # For more details about the formulation below, see section 4.3 in
            # Chen, V., Dunlop, M. M., Papaspiliopoulos, O., & Stuart, A. M.
            # (2018). Dimension-robust MCMC in Bayesian inverse problems.
            # arXiv preprint arXiv:1803.03344.
            field_list[:,idx] = self.eig_vec@( self.eig_val**((nu+d/2)/2)*p[...,idx] )

        if len(field_list) == 1:
            return field_list[0]
        else:
            return field_list

    def _build_basis(self):
        """Builds the basis of expansion of the Matern covariance operator"""
        # Define function space, test and trial functions
        V = self._build_space()
        u = dl.TrialFunction(V)
        v = dl.TestFunction(V)

        # Define the weak form a of the differential operator used in building the Matern field basis
        tau2 = 1/self.length_scale/self.length_scale
        a = tau2*u*v*dl.dx + dl.inner(dl.grad(u), dl.grad(v))*dl.dx
        
        # Set up the boundary conditions of the SPDE
        if self.boundary_conditions.lower() == 'neumann':
            boundary = lambda x, on_boundary: False
        elif self.boundary_conditions.lower() == 'zero':
            boundary = lambda x, on_boundary: on_boundary
        else:
            raise ValueError(f"Boundary conditions {self.boundary_conditions}, is not supported")

        u0 = dl.Constant('0.0')
        bc = dl.DirichletBC(V, u0, boundary)
        
        # Assemble the differential operator
        u_fun = dl.Function(V)
        L = u_fun*v*dl.dx
        K = dl.PETScMatrix()
        dl.assemble_system(a, L, bc, A_tensor=K)

        # Compute the first self.num_terms eigenvalues and eigenvectors of the 
        # (inverse) of the differential operator
        eigen_solver = dl.SLEPcEigenSolver(K)
        eigen_solver.parameters['spectrum'] = 'smallest magnitude'

        eigen_solver.solve(self.num_terms)
        self._eig_val = np.zeros(self.num_terms)
        self._eig_vec = np.zeros( [ u_fun.vector().get_local().shape[0], self.num_terms ] )

        for i in range( self.num_terms ):
            val, c, vec, cx = eigen_solver.get_eigenpair(i)
            self._eig_val[i] = val
            self._eig_vec[:,i] = vec.get_local()

        self._eig_val = np.reciprocal( self._eig_val )

        # Normalize the eigenvectors if required
        if self.normalize:
            self._eig_vec /= np.linalg.norm( self._eig_vec )


    def _build_space(self):
        """Create the function space on which the Matern covariance is discretized"""

        if hasattr(self.geometry, 'mesh'): 
            mesh = self.geometry.mesh
            V = dl.FunctionSpace(mesh, "CG", 1)
	
        else:
            raise NotImplementedError

        return V



import dolfin as dl
from numbers import Number

class ExpressionFromCallable(dl.UserExpression):
    """A `dolfin.Expression` that is created from a callable. The callable
    should take a single argument, the coordinate vector `x` and return a
    number, the value of the expression at `x`. The keyword arguments are passed
    to the `dolfin.Expression` constructor. An example of the keyword arguments 
    is the `degree` argument for `dolfin.Expression` which is used to specify 
    the element (polynomial) degree of the expression.
    """
    def __init__(self, func, **kwargs):
        super().__init__(**kwargs)
        self.func = func

    def eval(self, value, x):
        value[:] = self.func(x)


def to_dolfin_expression(value, **expression_kwargs):
    """Converts a value to a `dolfin.Expression`. The value should be a 
    callable, `dolfin.Expression` or a number. expression_kwargs are passed to
    the `dolfin.Expression` constructor if the value is a callable. An example 
    of the keyword arguments is the `degree` argument for `dolfin.Expression` 
    which is used to specify the element (polynomial) degree of the expression.
    """
    
    if not callable(value) and expression_kwargs != {}:
        raise ValueError("Cannot pass kwargs to non-callable value")
    if callable(value):
        return ExpressionFromCallable(value, **expression_kwargs)
    elif isinstance(value, dl.UserExpression) \
        or isinstance(value, dl.Expression)\
            or isinstance(value, dl.Constant):
        return value
    elif isinstance(value, Number):
        return dl.Constant(value)
    else:
        raise ValueError("Cannot convert to dolfin.Expression")




import dolfin as dl
from cuqipy_fenics.geometry import FEniCSContinuous, MaternKLExpansion
import numpy as np
import pytest


def test_MaternKLExpansion():
    """Test creating a MaternKLExpansion geometry"""
    mesh = dl.UnitSquareMesh(20, 20)
    V = dl.FunctionSpace(mesh, 'CG', 1)
    geometry = FEniCSContinuous(V)
    MaternGeometry = MaternKLExpansion(geometry,
                                     length_scale=.2,
                                     num_terms=128)
    assert (MaternGeometry.num_terms == 128 and np.isclose(
        MaternGeometry.length_scale, .2))


def test_MaternKLExpansion_basis(copy_reference):
    """Test MaternKLExpansion geometry basis building"""

    # Create the MaternKLExpansion geometry
    np.random.seed(0)
    mesh = dl.UnitSquareMesh(20, 20)
    V = dl.FunctionSpace(mesh, 'CG', 1)
    geometry = FEniCSContinuous(V)
    MaternGeometry = MaternKLExpansion(geometry,
                                     length_scale=.2,
                                     num_terms=128,
                                     normalize=False)

    # Build the basis
    MaternGeometry._build_basis()

    # Read the reference eigenvalues and eigenvectors
    samples_orig_file = copy_reference("data/MaternExpansion_basis.npz")
    expected_eig_pairs = np.load(samples_orig_file)
    expected_eig_vec = expected_eig_pairs['expected_eig_vec']
    expected_eig_val = expected_eig_pairs['expected_eig_val']

    # Assert that the eigenvalues match the reference
    assert (np.allclose(MaternGeometry._eig_val, expected_eig_val))

    # Assert that the eigenvectors match the reference, up to sign
    for i in range(MaternGeometry.num_terms):
        assert (np.allclose(MaternGeometry.eig_vec[:, i], expected_eig_vec[:, i]) or np.allclose(
            MaternGeometry.eig_vec[:, i], -expected_eig_vec[:, i]))

    # Assert that the eigenvectors has the correct shape
    assert expected_eig_vec.shape == (441, 128)


@pytest.mark.parametrize("nu, valid",
                         [(0, False), (0.01, True), (-1, False)])
def test_MaternKLExpansion_nu(nu, valid):
    """Test passing nu to the MaternKLExpansion geometry"""
    mesh = dl.UnitSquareMesh(20, 20)
    V = dl.FunctionSpace(mesh, 'CG', 1)
    geometry = FEniCSContinuous(V)

    if valid:
        MaternGeometry = MaternKLExpansion(geometry,
                                         length_scale=0.2,
                                         nu=nu,
                                         num_terms=128)
    else:
        with pytest.raises(ValueError):
            MaternGeometry = MaternKLExpansion(geometry,
                                             length_scale=0.2,
                                             nu=nu,
                                             num_terms=128)



import dolfin as dl
import cuqipy_fenics
import numpy as np
import pytest
import matplotlib.pyplot as plt
from cuqi.distribution import Gaussian
from cuqi.geometry import Continuous1D
from cuqipy_fenics.testproblem import FEniCSPoisson2D


@pytest.mark.parametrize("bc_types, bc_values, case", [
    (None, None, "valid_1"),
    (['Neumann', 'Dirichlet', 'dirichlet', 'neumann'],
     [-2, lambda x: 0.9*np.sin(x[0])+1, 1, lambda x: x[0]+0.1],
     "valid_2"),
    (['Dirichlet', 'Dirichlet', 'Dirichlet', 'Dirichlet'],
     [0, 3, 0, 1],
     "valid_3"),
    (['Neumann', 'Neumann', 'Neumann', 'neumann'],
     [0, 1, 0, 1],
     "all_neumann_error"),
    (['Neumann', 'Dirichlet'],
     [0, 1],
     "missing_bc_error")])
def test_FEniCSPoisson2D_bc(bc_types, bc_values, case):
    """Test creating a FEniCSPoisson2D testproblem with different boundary conditions"""
    if case[:5] == "valid":
        testproblem = FEniCSPoisson2D((20, 20),
                                      bc_types=bc_types, bc_values=bc_values, source_term=0)

        testproblem.data.plot()
        testproblem.exactData.plot()
        testproblem.exactSolution.plot()

    if case == "all_neumann_error":
        with pytest.raises(ValueError, match=r"All boundary conditions cannot be Neumann"):
            testproblem = FEniCSPoisson2D((20, 20),
                                          bc_types=bc_types, bc_values=bc_values, source_term=0)

    if case == "missing_bc_error":
        with pytest.raises(ValueError, match=r"length of bc_types list should be 4"):
            testproblem = FEniCSPoisson2D((20, 20),
                                          bc_types=bc_types, bc_values=bc_values, source_term=0)

    # Check that the Dirichlet boundary conditions are sat up
    #  correctly by comparing to the exact solution
    if case == "valid_3":
        func = testproblem.exactData.funvals
        assert np.isclose(func(0, 0.5), bc_values[0])
        assert np.isclose(func(0.5, 0), bc_values[1])
        assert np.isclose(func(1, 0.5), bc_values[2])
        assert np.isclose(func(0.5, 1), bc_values[3])

    if case == "valid_2":
        func = testproblem.exactData.funvals
        assert np.isclose(func(0.5, 0), bc_values[1]([0.5, 0]))
        assert np.isclose(func(1, 0.5), bc_values[2])


@pytest.mark.parametrize("source_term, valid", [
    (None, True),
    (3, True),
    (lambda x: x[0], True),
    (dl.Constant(4), True)])
def test_FEniCSPoisson2D_source_term(source_term, valid):
    """Test creating a FEniCSPoisson2D testproblem with different source terms"""
    if valid:
        testproblem = FEniCSPoisson2D((30, 30), source_term=source_term)

        testproblem.data.plot()
        testproblem.exactData.plot()
        testproblem.exactSolution.plot()


@pytest.mark.parametrize(
    "exactSolution, noise_level, field_type, field_params, mapping," +
    "prior, case",
    [(None, 0.02, None, None, None, None, "valid"),
     (None, 0.02, None, None, None, None, "unknown_field"),
     (None, 0.05, None, None, "exponential", Gaussian(0, np.ones(21*21)), "valid"),
     (lambda x: x[0]+x[1]+0.1, 0.02, None, None, "exponential",
      Gaussian(0, np.ones(21*21), name='x'), "valid"),
     (None, 0.05, None, None, lambda m: m+1, None, "valid"),
     (None, 0.1, "KL", None, "exponential", None, "valid"),
     (np.random.randn(10), 0.02, "KL",
      {'length_scale': 0.2, 'num_terms': 10}, "exponential", None, "valid")])
def test_FEniCSPoisson2D_setup(exactSolution, noise_level,
                               field_type, field_params, mapping,
                               prior, case):
    """Test creating a FEniCSPoisson2D testproblem with different 
    parametrization, exact solution and noise level"""
    if case == "valid":
        testproblem = FEniCSPoisson2D(
            (20, 20),
            exactSolution=exactSolution,
            noise_level=noise_level,
            field_type=field_type,
            field_params=field_params,
            mapping=mapping,
            prior=prior)

        testproblem.data.plot()
        testproblem.exactData.plot()
        testproblem.exactSolution.plot()

    if case == "unknown_field":
        with pytest.raises(ValueError, match=r"Unknown field type"):
            testproblem = FEniCSPoisson2D(
                (20, 20),
                exactSolution=exactSolution,
                noise_level=noise_level,
                field_type=Continuous1D(21*21),
                field_params=field_params,
                mapping=mapping,
                prior=prior)

    # Check the noise level is correct
    if case == "valid":
        noise_norm =\
            np.linalg.norm(testproblem.data.funvals.vector().get_local()
                           - testproblem.exactData.funvals.vector().get_local())
        exact_norm = np.linalg.norm(
            testproblem.exactData.funvals.vector().get_local())
        assert np.isclose(noise_norm/exact_norm, noise_level)



import sys
import numpy as np
import matplotlib.pyplot as plt
import cuqipy_fenics
import dolfin as dl
import warnings
import cuqi
from cuqi.problem import BayesianProblem
from cuqi.model import PDEModel
from cuqi.distribution import Gaussian
from cuqi.geometry import Geometry
import dolfin as dl
import ufl
from .pde import SteadyStateLinearFEniCSPDE
from .geometry import FEniCSContinuous, FEniCSMappedGeometry,\
      MaternKLExpansion
from .utilities import to_dolfin_expression


__all__ = ['FEniCSDiffusion1D', 'FEniCSPoisson2D']

class FEniCSDiffusion1D(BayesianProblem):
    """
    1D Diffusion PDE problem using FEniCS. The problem has Dirichlet boundary conditions.

    Parameters
    ------------
    dim : int, Default 100
        Dimension of the 1D problem

    endpoint : float, Default 1
        Endpoint of the 1D grid (starts at 0).

    exactSolution : ndarray, CUQIarray, Default None
        Exact solution used to generate data. 
        If None a default exact solution is chosen.

    SNR : float, Default 100
        Signal-to-noise ratio.
    
    mapping: str or callable
        mapping to parametrize the Bayesian parameters. If provided as string, it can take one of the values: ['exponential']

    Attributes
    ----------
    data : ndarray
        Generated (noisy) data

    model : cuqi.model.Model
        The forward model

    prior : cuqi.distribution.Distribution
        Distribution of the prior (Default = None)

    likelihood : cuqi.likelihood.Likelihood
        Likelihood function.

    exactSolution : ndarray
        Exact solution (ground truth)

    exactData : ndarray
        Noise free data

    infoSring : str
        String with information about the problem, noise etc.

    Methods
    ----------
    MAP()
        Compute MAP estimate of posterior.
        NB: Requires prior to be defined.

    sample_posterior(Ns)
        Sample Ns samples of the posterior.
        NB: Requires prior to be defined.

    """
    
    def __init__(self, dim = 100, endpoint = 1, exactSolution = None, SNR = 100, observation_operator = None, mapping = None, left_bc=0, right_bc=1):

        # Create FEniCSPDE        
        def u_boundary(x, on_boundary):
            return on_boundary

        def form(m,u,p):
            return m*ufl.inner(ufl.grad(u), ufl.grad(p))*ufl.dx
        
        mesh = dl.IntervalMesh(dim, 0,endpoint)

        solution_function_space = dl.FunctionSpace(mesh, 'Lagrange', 1)
        parameter_function_space = dl.FunctionSpace(mesh, 'Lagrange', 1)

        dirichlet_bc_expression = dl.Expression("left_bc*(x[0]<eps)+right_bc*(x[0]>endpoint-eps)", eps=dl.DOLFIN_EPS, endpoint=endpoint, left_bc=left_bc, right_bc=right_bc, degree=1)
        dirichlet_bc = dl.DirichletBC(solution_function_space, dirichlet_bc_expression, u_boundary)
        adjoint_dirichlet_bc = dl.DirichletBC(
            solution_function_space, dl.Constant(0), u_boundary)

        PDE = cuqipy_fenics.pde.SteadyStateLinearFEniCSPDE(
            form, mesh, solution_function_space, parameter_function_space, dirichlet_bc, adjoint_dirichlet_bc, observation_operator=observation_operator)
        
        # Create PDE model
        domain_geometry = cuqipy_fenics.geometry.FEniCSContinuous(parameter_function_space)
        if mapping is not None:
            if mapping == 'exponential':
                mapping = lambda x : ufl.exp(x)
            elif callable(mapping):
                mapping = mapping
            else:
                raise ValueError('mapping can be a callable or one of the expected keywords.')
            domain_geometry =  cuqipy_fenics.geometry.FEniCSMappedGeometry(geometry=domain_geometry, map = mapping)

        range_geometry = cuqipy_fenics.geometry.FEniCSContinuous(solution_function_space)
        
        model = cuqi.model.PDEModel(PDE,range_geometry,domain_geometry)

        # Create prior
        pr_mean = np.zeros(domain_geometry.par_dim)
        x = cuqi.distribution.GMRF(pr_mean,25,1,'zero') 
        
        # Set up exact solution
        if exactSolution is None:
            exactSolution = x.sample()
        elif exactSolution == 'smooth_step':
            N = dim + 1
            fun = lambda grid:  0.8*np.exp( -( (grid -endpoint/2.0 )**2 ) / 0.02)
            grid = np.linspace(0,endpoint,N)
            exactSolution = np.ones(N)*.8
            exactSolution[np.where(grid > endpoint/2.0)
                          ] = fun(grid[np.where(grid > endpoint/2.0)])
            exactSolution = cuqi.array.CUQIarray(
                exactSolution, geometry=domain_geometry)

        # Generate exact data
        b_exact = model.forward(domain_geometry.par2fun(exactSolution),is_par=False)

        # Add noise to data
        # Reference: Adding noise with a desired signal-to-noise ratio
        # https://sites.ualberta.ca/~msacchi/SNR_Def.pdf
        noise = np.random.normal(0, 1, b_exact.shape)
        alpha = np.linalg.norm(b_exact)/(np.sqrt(SNR)*np.linalg.norm(noise))

        data = cuqi.array.CUQIarray(
            b_exact + alpha*noise, geometry=range_geometry)

        # Create likelihood
        y = cuqi.distribution.Gaussian(
            mean=model(x), cov=alpha*np.eye(range_geometry.par_dim))

        # Initialize FEniCSDiffusion1D as BayesianProblem problem
        super().__init__(y, x, y=data)

        # Store exact values and information
        self.exactSolution = exactSolution
        self.exactData = b_exact
        self.infoString = f"Noise type: Additive i.i.d. noise with mean zero and signal to noise ratio: {SNR}"


class FEniCSPoisson2D(BayesianProblem):
    """
    2D Diffusion PDE-based Bayesian inverse problem that uses FEniCS. 
    The problem is set up on a unit square ([0,1]x[0,1]) mesh with either Dirichlet
    or Neumann boundary conditions on each boundary. The unknown parameter
    is the (possibly heterogeneous) diffusion coefficient (e.g. conductivity)
    field. The unknown parameter (e.g. conductivity) and the PDE solution 
    (e.g. the potential) are approximated in the first order Lagrange FEM space.

    Parameters
    -----------
    dim : tuple, Default (32,32)
        | Number of the 2D mesh vertices on the x and y directions, respectively.

    bc_type : list of str, Default ['Dirichlet', 'Dirichlet', 'Dirichlet', 'Dirichlet']
        | Boundary conditions on each boundary. The accepted values are:
        | 'Dirichlet': Dirichlet boundary condition.
        | 'Neumann': Neumann boundary condition.
        | The list should be ordered as follows: [left, bottom, right, top]

    bc_value : list of entries, each is a float or a callable , Default [0, 0, 0, 0]
        | Boundary condition values on each boundary. The accepted values are:
        | A float: a constant value.
        | A callable : a callable that takes a point coordinate vector `x` in the physical domain as input and return the boundary condition value at that point, e.g., `lambda x: np.sin(x[0])+np.cos(x[1])`.
        | The list should be ordered as follows: [left, bottom, right, top]

    exactSolution : ndarray, CUQIarray, or callable , Default None
        | Exact solution to the Bayesian inverse problem used to generate data, the diffusivity coefficient field in this case. When passed as a callable, it should take a point coordinate vector `x` in the physical domain as input and return the exact solution value at that point, e.g. `lambda x: np.sin(x[0])+np.cos(x[1])`. If None, a default exact solution is chosen. The default exact solution is a prior sample if the `field_type` is 'KL' and a smooth function if the `field_type` is None.

    source_term : float, callable, or dolfin.Expression, Default 1
        | Source term in the PDE. The accepted values are:
        | A float: a constant value.
        | A callable: a callable that takes a point coordinate vector `x` in the physical domain as input and returns the source term value at that point, e.g. `lambda x: np.sin(x[0])+np.cos(x[1])`.
        | A dolfin.Expression: a dolfin.Expression object that defines the source term.
        
    noise_level : float, default 0.01
        Noise level relative to the exact data (the ratio of the L2 norm of the noise to the L2 norm of the exact data). By default, the noise level is 1% (=0.01).

    field_type : str, Default None
        | Field type of the forward model domain. The accepted values are:
        | "KL": a :class:`MaternKLExpansion` geometry object will be created and set as a domain geometry.
        | None: a :class:`FEniCSContinuous` geometry object will be created and set as a domain geometry.

    field_params : dict, Default None
        | A dictionary of keyword arguments that the underlying geometry accepts. (Passed to the underlying geometry when field type is "KL" or None). For example, for "KL" field type, the dictionary can be `{"length_scale": 0.1, "num_terms": 32}`. If None is passed as field_type, this argument is ignored.

    mapping : str or callable , Default None
        | mapping to parametrize the Bayesian parameters. If None, no mapping is applied. If provided as callable, it should take a FEniCS function (of the unknown parameter) as input and return a FEniCS form, e.g. `lambda m: ufl.exp(m)`.
        | If provided as string, it can take one of the values: 
        | 'exponential' : Parameterization in which the unknown parameter becomes the log of the diffusion coefficient.

    prior : cuqi.distribution.Distribution, Default Gaussian
        | Distribution of the prior. Needs to be i.i.d standard Gaussian if field_type is "KL". The prior name property, i.e., `prior.name` is expected to be "x".
    """

    def __init__(self, dim=None, bc_types=None, bc_values=None,
                 exactSolution=None, source_term=None, noise_level=None,
                 field_type=None, field_params=None, mapping=None, prior=None):

        # Create the mesh
        if dim is None:
            dim = (32, 32)
        mesh = dl.UnitSquareMesh(dim[0], dim[1])

        # Create the function space
        V = dl.FunctionSpace(mesh, 'Lagrange', 1)

        # Set up boundary conditions
        if bc_types is None:
            bc_types = ['Dirichlet', 'Dirichlet', 'Dirichlet', 'Dirichlet']
        elif len(bc_types) != 4:
            raise ValueError(
                "The length of bc_types list should be 4. The list should be ordered as follows: [left, bottom, right, top]")
        elif all(bc_type.lower() in ['neumann'] for bc_type in bc_types):
            raise ValueError(
                "All boundary conditions cannot be Neumann. At least one boundary condition should be Dirichlet.")
        if bc_values is None:
            bc_values = [0, 0, 0, 0]
        elif len(bc_values) != 4:
            raise ValueError(
                "The length of bc_values list should be 4. The list should be ordered as follows: [left, bottom, right, top]")
        bc_values = [to_dolfin_expression(bc_value) for bc_value in bc_values]

        subdomains = self._create_boundaries_subdomains()
        # Set up Neumann and Dirichlet boundary conditions
        dirichlet_bcs = self._set_up_dirichlet_bcs(
            V, bc_types, bc_values, subdomains)
        neumann_bcs = self._set_up_neumann_bcs(
            V, bc_types, bc_values, subdomains)
        
        # Set up the adjoint boundary conditions used for solving 
        # the Poisson adjoint problem, which is needed for computing
        # the adjoint-based gradient.
        adjoint_dirichlet_bcs = self._set_up_adjoint_dirichlet_bcs(
            V, bc_types, subdomains)

        # Set up the source term
        if source_term is None:
            source_term = dl.Constant(1)
        else:
            source_term = to_dolfin_expression(source_term)

        # Set up the variational problem form
        if mapping is None:
            def parameter_form(m): return m
        elif callable(mapping):
            parameter_form = mapping
        elif mapping.lower() == 'exponential':
            def parameter_form(m): return ufl.exp(m)
        else:
            raise ValueError('mapping should be a callable, None or a string.')

        def form(m, u, p):
            return parameter_form(m)*ufl.inner(ufl.grad(u), ufl.grad(p))*ufl.dx\
                - source_term*p*ufl.dx\
                - neumann_bcs(m, p)

        # Create the CUQI PDE object
        PDE = SteadyStateLinearFEniCSPDE(
            form,
            mesh,
            parameter_function_space=V,
            solution_function_space=V,
            dirichlet_bcs=dirichlet_bcs,
            adjoint_dirichlet_bcs=adjoint_dirichlet_bcs)

        # Create the domain geometry
        G_FEM = FEniCSContinuous(V)
        if field_params is None:
            field_params = {}
        if field_type is None:
            G_domain = G_FEM
        elif field_type == 'KL':
            if field_params == {}:
                field_params = {'length_scale': 0.1, 'num_terms': 32}
            G_domain = MaternKLExpansion(G_FEM, **field_params)
        else:
            raise ValueError('Unknown field type.')

        # Create the range geometry
        G_range = FEniCSContinuous(V)

        # Create the forward model
        A = PDEModel(PDE, domain_geometry=G_domain, range_geometry=G_range)

        # Create the prior
        if prior is None:
            prior = Gaussian(np.zeros(A.domain_dim), 1,
                             geometry=G_domain, name='x')
            
        A = A(prior) # Set parameter name of model to match prior

        # Set up the exact solution
        if exactSolution is None and field_type == "KL":
            rng = np.random.RandomState(15)
            exactSolution = rng.randn(G_domain.par_dim)
        elif exactSolution is None:
            def exactSolution(x): return 1.5 + 0.5 * \
                np.sin(2*np.pi*x[0])*np.sin(2*np.pi*x[1])

        if isinstance(exactSolution, np.ndarray):
            exactSolution = cuqi.array.CUQIarray(
                exactSolution,
                is_par=True,
                geometry=G_domain)

        elif callable(exactSolution):
            exactSolution_expr = to_dolfin_expression(exactSolution)
            exactSolution_func = dl.interpolate(exactSolution_expr, V)
            exactSolution = cuqi.array.CUQIarray(
                exactSolution_func,
                is_par=False,
                geometry=G_domain)

        else:
            raise ValueError(
                'exactSolution should be a numpy array, a function or None.')

        # Create the exact data
        exact_data = A(exactSolution)
        if not isinstance(exact_data, cuqi.array.CUQIarray):
            exact_data = cuqi.array.CUQIarray(
                exact_data, is_par=True, geometry=G_range)

        # Create the data distribution and the noisy data
        noise = np.random.randn(len(exact_data))
        if noise_level is None:
            noise_level = 0.01
        noise_std = noise_level * \
            np.linalg.norm(exact_data)/np.linalg.norm(noise)
        noise = noise_std*noise
        data = exact_data + noise

        data_dist = Gaussian(mean=A(prior), cov=noise_std**2, geometry=G_range,
                             name='y')

        # Create the Bayesian problem
        super().__init__(data_dist, prior)
        super().set_data(y=data)

        # Store exact values and information
        self.exactSolution = exactSolution
        self.exactData = exact_data
        self.infoString = f"Noise type: Additive i.i.d. noise with standard deviation: {noise_std} and relative noise standard deviation: {noise_level}."

    def _create_boundaries_subdomains(self):
        """
        Create subdomains for the boundary conditions.
        """
        class Left(dl.SubDomain):
            def inside(self, x, on_boundary):
                return on_boundary and x[0] < dl.DOLFIN_EPS

        class Bottom(dl.SubDomain):
            def inside(self, x, on_boundary):
                return on_boundary and x[1] < dl.DOLFIN_EPS

        class Right(dl.SubDomain):
            def inside(self, x, on_boundary):
                return on_boundary and x[0] > 1.0 - dl.DOLFIN_EPS

        class Top(dl.SubDomain):
            def inside(self, x, on_boundary):
                return on_boundary and x[1] > 1.0 - dl.DOLFIN_EPS

        return [Left(), Bottom(), Right(), Top()]

    def _set_up_dirichlet_bcs(self, V, bc_types, bc_values, subdomains):
        """
        Set up Dirichlet boundary conditions for the Poisson PDE problem defined
        on the unit square mesh, where V is the function space.
        """
        dirichlet_bcs = []

        for i, bc in enumerate(bc_types):
            if bc.lower() == 'dirichlet':
                dirichlet_bcs.append(dl.DirichletBC(
                    V, bc_values[i], subdomains[i]))

        return dirichlet_bcs

    def _set_up_adjoint_dirichlet_bcs(self, V, bc_types, subdomains):
        """
        Set up Dirichlet boundary conditions for the adjoint Poisson PDE problem 
        defined on the unit square mesh, where V is the function space. The 
        adjoint problem is needed for the computation of the adjoint-based 
        gradient. The adjoint Dirichlet boundary conditions are set to zero in 
        this case.
        """
        adjoint_dirichlet_bcs = []

        for i, bc in enumerate(bc_types):
            if bc.lower() == 'dirichlet':
                adjoint_dirichlet_bcs.append(dl.DirichletBC(
                    V, dl.Constant(0), subdomains[i]))

        return adjoint_dirichlet_bcs

    def _set_up_neumann_bcs(self, V, bc_types, bc_values, subdomains):
        """
        Set up Neumann boundary conditions for the Poisson PDE problem defined
        on the unit square mesh, where V is the function space.
        """

        boundary_markers = dl.MeshFunction(
            'size_t', V.mesh(), V.mesh().topology().dim()-1)
        boundary_markers.set_all(0)
        for i, subdomain in enumerate(subdomains):
            subdomain.mark(boundary_markers, i+1)
        ds = dl.Measure('ds', domain=V.mesh(), subdomain_data=boundary_markers)

        neumann_bcs = []
        for i, bc_type in enumerate(bc_types):
            if bc_type.lower() == 'neumann':
                neumann_bcs.append(lambda m, p: bc_values[i]*p*ds(i+1))

        if neumann_bcs == []:
            return lambda m, p: dl.Constant(0)*p*dl.ds
        else:
            return lambda m, p: sum([nbc(m, p) for nbc in neumann_bcs])



import dolfin as dl
import cuqi
import cuqipy_fenics
import numpy as np
import pytest
import time
import ufl
import sys

def test_model_input():
    """Test passing different data structures for PDEModel input"""
    model = cuqipy_fenics.testproblem.FEniCSDiffusion1D().model
    V = model.pde.parameter_function_space

    # Test passing a CUQIarray containing parameters
    u = dl.Function(V)
    u_CUQIarray = cuqi.array.CUQIarray(u.vector().get_local(), is_par=True, geometry=model.domain_geometry)
    y = model(u_CUQIarray)

    # Test passing parameters as a numpy array
    u = dl.Function(V)
    u_numpy = u.vector().get_local()
    y = model(u_numpy, is_par=True)

    # Test passing a CUQIarray containing dolfin function
    u = dl.Function(V)
    u_CUQIarray = cuqi.array.CUQIarray(u, is_par=False, geometry=model.domain_geometry)
    y = model(u_CUQIarray)

    # Test passing a CUQIarray containing dolfin function wrapped in np.array
    # This is not recommended, but should work
    u = dl.Function(V)
    u_CUQIarray = cuqi.array.CUQIarray(np.array(u, dtype='O'), is_par=False, geometry=model.domain_geometry)
    y = model(u_CUQIarray)  

    # Test passing a dolfin function (should fail)
    u = dl.Function(V)
    with pytest.raises(AttributeError):
        y = model(u)


def test_solver_choice():
    """Test passing different solvers to PDEModel"""
    # Create the variational problem
    poisson = Poisson()

    # Set up model parameters
    m = dl.Function(poisson.parameter_function_space)
    m.vector()[:] = 1.0

    # Create a PDE object with Krylov solver
    PDE_with_solver1 = cuqipy_fenics.pde.SteadyStateLinearFEniCSPDE(
        poisson.form,
        poisson.mesh,
        poisson.solution_function_space,
        poisson.parameter_function_space,
        poisson.bcs,
        linalg_solve=dl.KrylovSolver())

    # Solve the PDE
    PDE_with_solver1.assemble(m)
    u1, info = PDE_with_solver1.solve()

    # Create a PDE object with direct LU solver
    PDE_with_solver2 = cuqipy_fenics.pde.SteadyStateLinearFEniCSPDE(
        poisson.form,
        poisson.mesh,
        poisson.solution_function_space,
        poisson.parameter_function_space,
        poisson.bcs,
        linalg_solve=dl.LUSolver())

    # Solve the PDE
    PDE_with_solver2.assemble(m)
    u2, info = PDE_with_solver2.solve()

    # Check that the solutions are the same
    assert np.allclose(u1.vector().get_local(), u2.vector().get_local())


def test_reuse_assembled():
    """Test that reusing the assembled and factored lhs gives the same solution
     and a better performance"""
    # Create the variational problem
    poisson = Poisson()
    poisson.mesh = dl.UnitSquareMesh(50, 50)

    # Set up model parameters
    m = dl.Function(poisson.parameter_function_space)
    m.vector()[:] = 1.0

    # Create a PDE object, generate an error if the full form is provided
    # but reuse_assembled is True
    with pytest.raises(ValueError):
        PDE = cuqipy_fenics.pde.SteadyStateLinearFEniCSPDE(
            poisson.form,
            poisson.mesh,
            poisson.solution_function_space,
            poisson.parameter_function_space,
            poisson.bcs,
            reuse_assembled=True)

    # Create a PDE object
    PDE = cuqipy_fenics.pde.SteadyStateLinearFEniCSPDE(
        (poisson.lhs_form, poisson.rhs_form),
        poisson.mesh,
        poisson.solution_function_space,
        poisson.parameter_function_space,
        poisson.bcs,
        reuse_assembled=True)

    # Solve the PDE
    PDE.assemble(m)
    u1, info = PDE.solve()

    # update the parameter and solve again
    sin_func = dl.Expression("sin(2*pi*x[0])", degree=1)
    m.interpolate(sin_func)
    t0 = time.time()
    PDE.assemble(m)
    u1, info = PDE.solve()
    t1 = time.time()
    t_first = t1 - t0

    # Solve the PDE again
    t0 = time.time()
    PDE.assemble(m)
    u2, info = PDE.solve()
    t1 = time.time()
    t_reuse_factors = t1 - t0

    # Check that the solutions are the same
    assert np.allclose(u1.vector().get_local(), u2.vector().get_local())

    # Check that using the reuse_assembled option is faster
    assert t_reuse_factors < 0.25*t_first


def test_form():
    """Test creating PDEModel with full form, and with lhs and rhs forms"""
    # Create the variational problem
    poisson = Poisson()

    # Set up model parameters
    m = dl.Function(poisson.parameter_function_space)
    m.vector()[:] = 1.0

    # Create a PDE object with full form
    PDE_with_full_form = cuqipy_fenics.pde.SteadyStateLinearFEniCSPDE(
        poisson.form,
        poisson.mesh,
        poisson.solution_function_space,
        poisson.parameter_function_space,
        poisson.bcs)

    # Solve the PDE
    PDE_with_full_form.assemble(m)
    u1, info = PDE_with_full_form.solve()

    # Create a PDE object with lhs and rhs forms
    PDE_with_lhs_rhs_forms = cuqipy_fenics.pde.SteadyStateLinearFEniCSPDE(
        (poisson.lhs_form, poisson.rhs_form),
        poisson.mesh,
        poisson.solution_function_space,
        poisson.parameter_function_space,
        poisson.bcs)

    # Solve the PDE
    PDE_with_lhs_rhs_forms.assemble(m)
    u2, info = PDE_with_lhs_rhs_forms.solve()

    # Check that the solutions are the same
    assert np.allclose(u1.vector().get_local(), u2.vector().get_local())

@pytest.mark.parametrize("case", [1, 2])
def test_with_updated_rhs(copy_reference, case):
    """Regression test for using the method with_updated_rhs and sharing the
    factored lhs matrix. Case 1: compare the solution reusing the factorization
    and without reusing the factorization. Case 2: compare the solution with
    results from the previous version of the code."""

    # Skip case 2 test if the operating system is not Mac OS X
    if  case == 2 and not sys.platform.startswith('darwin'):
        pytest.skip("Test on MAC OS only")

    # Set up first poisson problem
    poisson1 = Poisson()
    poisson1.mesh = dl.UnitSquareMesh(40, 40)
    poisson1.source_term = dl.Constant(1.0)

    # Set up second poisson problem
    poisson2 = Poisson()
    poisson2.mesh = poisson1.mesh
    poisson2.source_term = dl.Expression("sin(2*x[0]*pi)*sin(2*x[1]*pi)", degree=1)

    # Set up boundary function (where the Dirichlet boundary conditions are applied)
    def u_boundary(x, on_boundary):
        return on_boundary and\
            (x[0] < dl.DOLFIN_EPS or x[0] > 1.0 - dl.DOLFIN_EPS)

    poisson1.bcs = dl.DirichletBC(poisson1.solution_function_space,
                                 dl.Constant(0.0), u_boundary)
    poisson2.bcs = poisson1.bcs

    # Create two PDE objects with different rhs terms
    PDE1 = cuqipy_fenics.pde.SteadyStateLinearFEniCSPDE(
        (poisson1.lhs_form, poisson1.rhs_form),
        poisson1.mesh,
        parameter_function_space=poisson1.parameter_function_space,
        solution_function_space=poisson1.solution_function_space,
        dirichlet_bcs=poisson1.bcs,
        observation_operator=None,
        reuse_assembled=False)

    PDE2 = cuqipy_fenics.pde.SteadyStateLinearFEniCSPDE(
        (poisson2.lhs_form, poisson2.rhs_form),
        poisson2.mesh,
        parameter_function_space=poisson2.parameter_function_space,
        solution_function_space=poisson2.solution_function_space,
        dirichlet_bcs=poisson2.bcs,
        observation_operator=None,
        reuse_assembled=False)

    # Create domain geometry
    fenics_continuous_geo = cuqipy_fenics.geometry.FEniCSContinuous(
        poisson1.parameter_function_space)
    domain_geometry = cuqipy_fenics.geometry.MaternKLExpansion(
        fenics_continuous_geo, length_scale=.1, num_terms=5)

    # Create range geometry
    range_geometry = cuqipy_fenics.geometry.FEniCSContinuous(
        poisson1.solution_function_space)

    # Create cuqi forward model (two models corresponding to two PDE objects)
    cuqi_model1 = cuqi.model.PDEModel(
        PDE1, domain_geometry=domain_geometry, range_geometry=range_geometry)

    cuqi_model2 = cuqi.model.PDEModel(
        PDE2, domain_geometry=domain_geometry, range_geometry=range_geometry)

    # Create the prior
    x = cuqi.distribution.Gaussian(mean=np.zeros(cuqi_model1.domain_dim),
                                   cov=1, geometry=domain_geometry)

    # Create exact solution and data
    np.random.seed(0)
    exact_solution = cuqi.array.CUQIarray(
        np.random.randn(domain_geometry.par_dim),
        is_par=True, geometry=domain_geometry)

    data1 = cuqi_model1(exact_solution)
    data2 = cuqi_model2(exact_solution)

    # Create likelihood 1
    y1 = cuqi.distribution.Gaussian(mean=cuqi_model1(x),
                                    cov=np.ones(cuqi_model1.range_dim)*.01**2)
    y1 = y1(y1=data1)

    # Create likelihood 2
    y2 = cuqi.distribution.Gaussian(mean=cuqi_model2(x),
                                    cov=np.ones(cuqi_model2.range_dim)*.01**2)
    y2 = y2(y2=data2)

    # Create posterior
    cuqi_posterior = cuqi.distribution.JointDistribution(
        y1, y2, x)._as_stacked()

    # Solve the Bayesian problem
    # Sample the posterior (Case 1: no reuse of assembled operators)
    Ns = 20
    np.random.seed(0)
    sampler = cuqi.sampler.MH(cuqi_posterior)
    t0 = time.time()
    samples1 = sampler.sample_adapt(Ns, Nb=10)
    t1 = time.time()
    t_no_reuse = t1-t0
    samples1.geometry = domain_geometry

    if case == 1:
        # Set PDE2 to be a shallow copy of PDE1 but with different rhs
        PDE1.reuse_assembled = True
        PDE2 = PDE1.with_updated_rhs(poisson2.rhs_form)
        cuqi_model2.pde = PDE2

        # Sample the posterior again (Case 2: reuse of assembled operators)
        np.random.seed(0)
        sampler = cuqi.sampler.MH(cuqi_posterior)
        t0 = time.time()
        samples2 = sampler.sample_adapt(Ns, Nb=10)
        t1 = time.time()
        t_reuse = t1-t0

        # Check that the samples are the same
        assert np.allclose(samples1.samples, samples2.samples)
        
        # Check that reusing factorization and with_updated_rhs is faster
        assert t_reuse < 0.7*t_no_reuse

    elif case == 2:
        # Check that the samples matches the ones generated
        # before updating the library code to add the reuse_assembled functionality
        samples_orig = np.load(
            copy_reference('data/samples_test_with_rhs_write_from_pytests'+\
            '_e349834fc189e1e1da6a962cdfe449f94486824e.npz'))["samples1"]
        assert np.allclose(samples1.samples[1, :], samples_orig[1, :])


class Poisson:
    """Define the variational PDE problem for the Poisson equation in two
    ways: as a full form, and as lhs and rhs forms"""

    def __init__(self):

        # Create the mesh and define function spaces for the solution and the
        # parameter
        self.mesh = dl.UnitIntervalMesh(10)

        # Define the boundary condition
        self.bc_value = dl.Constant(0.0)

        # the source term
        self.source_term = dl.Expression('x[0]', degree=1)

    @property
    def form(self):
        return lambda m, u, v:\
            ufl.exp(m)*ufl.inner(ufl.grad(u), ufl.grad(v))*ufl.dx\
            + self.source_term*v*ufl.dx

    @property
    def lhs_form(self):
        return lambda m, u, v:\
            ufl.exp(m)*ufl.inner(ufl.grad(u), ufl.grad(v))*ufl.dx

    @property
    def rhs_form(self):
        return lambda m, v: -self.source_term*v*ufl.dx

    @property
    def solution_function_space(self):
        return dl.FunctionSpace(self.mesh, "Lagrange", 2)

    @property
    def parameter_function_space(self):
        return dl.FunctionSpace(self.mesh, "Lagrange", 1)

    @property
    def bcs(self):
        if not hasattr(self, "_bcs") or self._bcs is None:
            self._bcs = dl.DirichletBC(self.solution_function_space,
                                      self.bc_value, "on_boundary")
        return self._bcs

    @bcs.setter
    def bcs(self, bcs):
        self._bcs = bcs


def test_observation_operator_setter():
    """Test that the observation setter works as expected"""
    # Create the variational problem
    poisson = Poisson()
    poisson.mesh = dl.UnitSquareMesh(50, 50)

    # Set up model parameters
    m = dl.Function(poisson.parameter_function_space)
    m.vector()[:] = 1.0

    # Create the observation operator
    observation_operator = lambda m, u: u.vector().get_local()[200:205] 

    # Create a PDE object (set observation_operator through setter)
    PDE1 = cuqipy_fenics.pde.SteadyStateLinearFEniCSPDE(
        (poisson.lhs_form, poisson.rhs_form),
        poisson.mesh,
        poisson.solution_function_space,
        poisson.parameter_function_space,
        poisson.bcs,
        observation_operator=None)
    PDE1.observation_operator = observation_operator

    # Create a PDE object (pass observation_operator as argument)  
    PDE2 = cuqipy_fenics.pde.SteadyStateLinearFEniCSPDE(
        (poisson.lhs_form, poisson.rhs_form),
        poisson.mesh,
        poisson.solution_function_space,
        poisson.parameter_function_space,
        poisson.bcs,
        observation_operator=observation_operator)

    # Solve the first PDE
    PDE1.assemble(m)
    u1, info = PDE1.solve()
    u1_obs = PDE1.observe(u1)

    # Solve the second PDE
    PDE2.assemble(m)
    u2, info = PDE2.solve()
    u2_obs = PDE2.observe(u2)

    # Apply the observation operator directly
    u3_obs = observation_operator(None, u2)

    # Check that the solutions are the same
    assert np.allclose(u1_obs, u2_obs) and np.allclose(u2_obs, u3_obs)
    assert len(u1_obs) == 5
